Program Aniwemva_steepds_rand

  use sep

  ! Aniso modules
  use ani_type
  use ani_parms_cg
  use ani_exec_time
  use ani_taper
  use ani_steepds_rand
  use ani_imgprc

  ! Lloyd referene model building
  use lloyd_model

  implicit none

  real						        					:: start,finish,start1,finish1
  real, allocatable										:: refr(:,:,:), taper(:)
  real, allocatable                                     :: dmodel(:), temp_model(:,:,:), nmodel(:,:,:), model(:,:,:)
  real, allocatable                                     :: B_image(:,:,:), D_image(:,:,:), t_image(:,:,:)
  complex, allocatable       			     			:: source1(:)
  integer					       						:: sh_position, iter
  integer                                               :: iz, ix, isecdim, isdso, isimage
  real, allocatable                                     :: l(:,:,:), u(:,:,:)
  double precision, allocatable                         :: fun(:)
  double precision                                      :: err, funt
  real                                                  :: lamda, lam_min, factor, tola
  real                                                  :: p_hi(2), p_lo(2)
  integer, allocatable                                  :: nstart(:)
  integer												:: length, seekpos
  integer, external										:: sreed, srite, sseek


  call sep_init()

  call aniparam()

  write(0,*) "---------------------------------------------------------------"
  write(0,*) "                    Initialization                             "
  write(0,*) "---------------------------------------------------------------"

  call from_param("isdso", isdso, 0)
  call from_param("isimage", isimage, 0)
  start=0.;finish=0.
  call aniexec_time(start)

  ! Backgournd model, in velocity 
  allocate(model(nx,nz,npar_in))
  if( mdl%iosize .ne.  sreed("mdl", model, mdl%iosize) ) then
		  call seperr("Input model size .ne. size of model")
  end if

  ! Mysterious negative eta and dlt value
  where(model(:,:,2)<1e-6) model(:,:,2) = 0.
  where(model(:,:,3)<1e-6) model(:,:,3) = 0.
  write(0,*) "initial velocity ", maxval(model(:,:,1)), minval(model(:,:,1))
  write(0,*) "initial eta ",   maxval(model(:,:,2)), minval(model(:,:,2))
  if (npar_in==3) write(0,*) " exist delta ",   maxval(model(:,:,3)), minval(model(:,:,3))

      ! Space for new model
      allocate(nmodel(nx,nz,npar_in)); nmodel = model;

      ! Space for model updates
      allocate(dmodel(nx*nz*npar_inv)); dmodel=0.

      ! Space for convert 1D dmodel to 3D dmodel
      allocate(temp_model(nx,nz,npar_inv))

      ! Bounds for model updates; bounds are converted to slowness
      allocate(l(nx,nz,npar_in)); l=0.
      allocate(u(nx,nz,npar_in)); u=0.
      call from_param('p_hi', p_hi)
      call from_param('p_lo', p_lo)
      u(:,:,1) = model(:,:,1) * (1-p_hi(1)); write(0,*) "upper bound for vel", maxval(u(:,:,1)) ; 
      l(:,:,1) = model(:,:,1) * (1+p_lo(1)); write(0,*) "lower bound for vel", minval(l(:,:,1)) ;
      u(:,:,2) = p_hi(2) ; write(0,*) "upper bound for eta", maxval(u(:,:,2)) 
      l(:,:,2) = -p_lo(2)  ; write(0,*) "lower bound for eta", minval(l(:,:,2)) 
      u(:,:,3) = model(:,:,3); write(0,*) "upper bound for dlt", maxval(u(:,:,3)) 
      l(:,:,3) = model(:,:,3); write(0,*) "upper bound for dlt", minval(l(:,:,3)) 

      ! nstart
      allocate(nstart(npar_in))
      call from_param("nstart", nstart)

      ! Convert everything to slowness
      model(:,:,1) =  1./model(:,:,1)   ! slowness; eta
      write(0,*) "after conver to slowness", maxval(model(:,:,1)), minval(model(:,:,1))
      u(:,:,1) = 1./u(:,:,1) ; write(0,*) "upper bound for slow", maxval(u(:,:,1)) ;
      l(:,:,1) = 1./l(:,:,1) ; write(0,*) "upper bound for slow", minval(l(:,:,1)) ;

      ! Reference model space
      allocate(refr(nref_max,nz,npar_in))

      ! Background image, perturbed image and temp image
      allocate(B_image(img%xm%n,nsecdim,img%z%n)); B_image=0.
      allocate(D_image(img%xm%n,nsecdim,img%z%n)); D_image=0.
      allocate(t_image(img%xm%n,nsecdim,img%z%n)); t_image=0.

      ! Source wavelet
      allocate(source1(sou%w%n)); 
	  if (sou%iosize .ne. sreed("sou", source1, sou%iosize) ) then
			  call seperr('Input size of Source .ne. size of source')
	  end if	

      ! Taper
      allocate(taper(nx)); call anitaper(tapertr,taper)

      ! Objective fun as iteration 
      allocate(fun(niter+1)); fun=0.;

      ! Steepest descent
      call from_param("factor", factor, 0.1)
      call from_param("lam_min", lam_min, 5.e-2)
      call from_param("tola", tola, 1.e-13)

      call aniexec_time(finish)
      write(0,'(8x,a,f10.2,a)')' Init.phase - Total elapsed time:', (finish - start),' (sec)'
      ! ---------------------------------------------------------------

      ! ---------------------------------------------------------------
      ! The main loop
      ! ---------------------------------------------------------------
      call steepds_init(source1, taper)  

      iter = 1
      err = 1

      do while (err > tola .and. iter <= niter)

        call fgeval_init(iter)

        write(0,*) "**************************************************"
        write(0,*) "         New iteration = ", iter
        write(0,*) "**************************************************"

        ! Get the backgound image for iter i
        if ( isdso==0 .or. ifec_img .or. isimage) then
          refr = 0.; call lloydmodel(scales, nstart, model, refr)
          call feval(model_in=model, refr_in=refr, Bimage=B_image)
          write(0,*) "Iter=", iter, "write the background image 1"
          if ( img%iosize .ne. srite("bgimage", B_image, img%iosize)) then 
				  call seperr('Ouput Size of img .ne. size of img')
		  end if
        else 
		  call aniwei_init(iter)
          seekpos = sseek("bgimage", -1*img%iosize, 2)
          write(0,*) "Iter=", iter, "read the background image", seekpos
          if ( img%iosize .ne. sreed("bgimage", B_image, img%iosize)) then
				  call seperr('Inpute size of img .ne. size of img')
		  end if		  
        end if
        ! Compute the perturbed image using dso
		call aniimgprc(.false.,.false.,t_image,B_image,mdso,fun(iter))
		write(0,*) "fun=",fun(iter), iter
		call aniimgprc(.false.,.true.,t_image,D_image,mdso)
        if ( img%iosize .ne. srite("dimage", D_image, img%iosize)) then
				call seperr("Output size of Dimage .ne. size of img")
		end if
        write(0,*) "after write the Dimage"
		if (isimage==1 .and. isdso==0 ) stop 

        ! Compute the gradient
        refr = 0.; call lloydmodel(scales, nstart, model, refr)
        dmodel=0.; call geval(model_in=model, refr_in=refr, Bimage=D_image, grad=dmodel)

!        if (iter==1 .and. isdso==1) stop

        ! Line Search  
        lamda = 0.5 ! /iter
        temp_model = reshape(dmodel,(/nx,nz,npar_inv/))
        nmodel(:,:,1:2) = model(:,:,1:2) + lamda * temp_model

        ! Hard bounds for the new model 
        where(nmodel < l);
          nmodel = l;
        end where
        where(nmodel > u);
          nmodel = u;
        end where
        write(0,*) "new slowness ", maxval(nmodel(:,:,1)), minval(nmodel(:,:,1))
        write(0,*) "new eta ",   maxval(nmodel(:,:,2)), minval(nmodel(:,:,2))
        write(0,*) "new delta ",   maxval(nmodel(:,:,3)), minval(nmodel(:,:,3))

        refr=0.; call lloydmodel(scales, nstart, nmodel, refr)
        call feval(model_in=nmodel, refr_in=refr, Bimage=B_image)
		call aniimgprc(.false.,.false.,t_image,B_image,mdso,funt)
		
        write(0,*) "1) fun new = ", sngl(funt), "fun = ", sngl(fun(iter)), "lamda=", lamda

        do while(funt > fun(iter)  .and. lamda > lam_min)
          lamda = lamda * factor
          nmodel(:,:,1:2) = model(:,:,1:2) + lamda * temp_model 
          where(nmodel < l);
            nmodel = l;
          end where
          where(nmodel > u);
            nmodel = u;
          end where
          refr=0.; call lloydmodel(scales, nstart, nmodel, refr)
          call feval(model_in=nmodel, refr_in=refr, Bimage=B_image)
		  call aniimgprc(.false.,.false.,t_image,B_image,mdso,funt)
          write(0,*) "2) fun new=", sngl(funt), "fun kept=", sngl(fun(iter)), "lamda=", lamda
        end do
        ! ******************** 
        ! End of Line Search
        ! ******************** 
		if (.not. ifec_img) then
		        call to_history("n4", iter, "bgimage")
				write(0,*) "Iter=", iter, "write the background image 2"
				if ( img%iosize .ne. srite("bgimage", B_image, img%iosize) ) then
						call seperr("Output size of Bimage .ne. size of img")
				end if
		end if

        model = nmodel 
        write(0,*) "new slowness ", maxval(model(:,:,1)), minval(model(:,:,1))
        write(0,*) "new eta ",   maxval(model(:,:,2)), minval(model(:,:,2))
        write(0,*) "new delta ",   maxval(model(:,:,3)), minval(model(:,:,3))


        call to_history("n4", iter)
        write(0,*) "Iter=", iter, "write the new model"
        if ( mdl%iosize .ne. srite("out", model, mdl%iosize) ) then
				call seperr("Output model size .ne. size of model")
		end if

        err  = sum(lamda*lamda*dmodel*dmodel) 
        tola = sum(nmodel(:,:,1:2)*nmodel(:,:,1:2)) * 1.e-10 
        write(0,*) 'Iteration = ',iter,' error=',err, ' tola=', tola
        write(0,*) "03) fn = ",funt
        iter = iter + 1
		isdso = 1
      end do

      call to_history("n1", niter, "fun")
      call to_history("d1",    1., "fun")
      call to_history("01",    1., "fun")
      length = srite("fun", sngl(fun), 4*niter)

      01  format(a9,i7,a6,i9)
      02  format(a9,i7,2(a6,i9))
      03  format(a9,i7)
      04  format(a12,i4,a7,i11)

      deallocate(model, nmodel, dmodel, temp_model)
      deallocate(source1, taper, refr)
      deallocate(l, u, fun) 
      deallocate(B_image)
      call sep_close()
end Program Aniwemva_steepds_rand

