module ani_parms_cg

  use sep
  use ani_type

  implicit none

  ! Forward or adjoint, dottest or not
  logical,        public, save				:: adj,dot

  ! Define the parameter grid given by input data
  type (dat_grid), public, save	        		:: rec, sou
  type (img_grid), public, save	        		:: img
  type (mdl_grid) , public, save				:: mdl
  type (axis) , public, save     			:: h, w, xd, xm, z, sh, ksh, th, par

  ! Define the parameters for spline smoothing
  type (axis) , public, save     			:: axk_v, azk_v     
  type (axis) , public, save     			:: axk_e, azk_e     

  ! Define parameters for triangle smoothing
  integer	 ,public, save				:: rect1_v, rect2_v, nsmth_v 
  integer	 ,public, save				:: rect1_e, rect2_e, nsmth_e 

  ! Define dip field for steering smoothing
  real, allocatable, public, save		:: dip(:)
  integer	, public, save				:: nang, nfilt, lscale, destroy
  real		, public, save				:: eps_steer

  ! Define mask for muting the water column
  real, allocatable, public, save		:: mute(:)

  ! Define the useful parameters in the program
  type (axis),    public, save  			:: kxm
  integer	 ,public, save				:: tapertr,padtraces
  integer	 ,public, save				:: node,nsecdim,mdso,niter, iter_v
  integer	 ,public, save				:: nx, nz, nt, nsh, npar_in, npar_inv, nref_max
  real   	 ,public, save				:: dx, ox, dz, oz, dt, ot, dsh, osh
  real           ,public, save				:: pi, osecdim, dsecdim, alpha
  real           ,public, save				:: illu_min, illu_max 
  real,    allocatable, public, save                    :: scales(:), g_scales(:)

  ! Define the encoding parameters
  logical	, public, save				:: ifec_img, ifec_grd
  integer	, public, save				:: rect1
  character	, public, save				:: cgtype*2
  character	, public, save				:: grdsm_style*4

  ! Some tempary parameters		
  integer								:: ec_img, ec_grd, gradiosize, ec_style
  integer								:: n1, n2
  real									:: minx, maxx
  contains

  subroutine aniparam()

    integer						:: iw, i

    ! ----------------------------------------------------------------------------------------------------------------------
    ! Read in parameters
    call from_param("dot",dot,.false.)
    call from_param("adj",adj,.true.)

	! Smoothing parameters for gradient field
	call from_param("rect1_v", rect1_v, 20)
	call from_param("rect2_v", rect2_v, 20)
	call from_param("rect1_e", rect1_e, 20)
	call from_param("rect2_e", rect2_e, 20)
	call from_param("nsmth_v", nsmth_v , 1)
	call from_param("nsmth_e", nsmth_e , 1)

	! encoding parameters
	call from_param("rect1_img", rect1, 10)   ! smoothing parameter for image
	call from_param("ec_img",ec_img,0); 
	call from_param("ec_grd",ec_grd,1)
	call from_param("ec_stl",ec_style,0)		! 0 - stationary encoding
												! 1 - linear increasing encoding
	if (ec_img==0) then 
			ifec_img=.false. 
    else 
			ifec_img=.true. 
    end if 
	if (ec_grd==0) then 
			ifec_grd=.false. 
    else 
			ifec_grd=.true. 
    end if 

	! Conjugate gradient parameter
	call from_param("cgtype", cgtype, "SD")    ! Default is steepest decent

    call from_param("node",node,9999)
    call from_param("traces_in_taper",tapertr,0)
    call from_param("padtraces",padtraces,tapertr)

    call from_param("niter", niter, 5)
    call from_param("iter_v", iter_v, 10)

    call from_param("illu_min",illu_min,  1.)
    call from_param("illu_max",illu_max, 50.)

    ! initialize file headers
    call auxin("rec")
    call auxin("sou")
    call auxin("mdl")
    call auxin("mask")

    ! Read the receiver parameter. xd is the shot position
    call from_aux("rec","n1",rec%h%n);	call from_aux("rec","d1",rec%h%d);call from_aux("rec","o1",rec%h%o)  ! offset axis
    call from_aux("rec","n2",rec%w%n);	call from_aux("rec","d2",rec%w%d);call from_aux("rec","o2",rec%w%o)  ! freq   axis
    call from_aux("rec","n3",rec%xd%n);call from_aux("rec","d3",rec%xd%d);call from_aux("rec","o3",rec%xd%o) ! shot index 

    ! Read the source parameter. 
    call from_aux("sou","n1",sou%w%n);	call from_aux("sou","d1",sou%w%d);	call from_aux("sou","o1",sou%w%o)

    ! Read the model parameters.
    call from_aux("mdl","n1",mdl%xm%n);	call from_aux("mdl","d1",mdl%xm%d);	call from_aux("mdl","o1",mdl%xm%o)
    call from_aux("mdl","n2",mdl%z%n );	call from_aux("mdl","d2",mdl%z%d );	call from_aux("mdl","o2",mdl%z%o )
    call from_aux("mdl","n3",mdl%par%n);	

    ! Read the parameters for computing reference model
    npar_in = mdl%par%n   ! # of par in the model

    if (mdl%par%n==3) then
      npar_inv = mdl%par%n-1;    ! # of par to invert
      write(0,*) "parameter # =", mdl%par%n
    else
      npar_inv = mdl%par%n
      write(0,*) "parameter # =", mdl%par%n
    end if

    call from_param("nref_max", nref_max, 3)
    allocate(scales(npar_in),g_scales(npar_inv))
    call from_param("scales", scales)
    write(0,*) scales
    call from_param("g_scales", g_scales)
    write(0,*) g_scales

    ! ----------------------------------------------------------------------------------------------------------------------
    ! Check the coherence of the source and receiver
    if (sou%w%n .ne. rec%w%n) then
      call seperr("# source freq. .ne. # receiver freq.")
    end if
    if (sou%w%d .ne. rec%w%d) then
      call seperr("source freq.interval .ne. receiver freq.interval")
    end if
    if (sou%w%o .ne. rec%w%o) then
      call seperr("source freq.origin .ne. receiver freq.origin")
    end if

    ! ----------------------------------------------------------------------------------------------------------------------
    ! Compute the useful parameters
    pi=4*atan(1.)
    nsh=rec%xd%n; dsh=rec%xd%d; osh=rec%xd%o

    ! ----------------------------------------------------------------------------------------------------------------------
    ! The final 2-D grid for wavefield and models 
    nz = mdl%z%n; dz = mdl%z%d; oz = mdl%z%o
    dx = mdl%xm%d
    minx = osh + rec%h%o-(padtraces+tapertr)*dx 
	maxx = osh + (nsh-1) * dsh + rec%h%o + (rec%h%n-1) * rec%h%d + (padtraces+tapertr)*dx
    nx = (maxx-minx)/mdl%xm%d + 1   
	ox = minx

	if (nx .ne. mdl%xm%n ) then
      write(0,*) "nx=", nx, "mdl%xm%n=", mdl%xm%n
      call seperr("# of working nx .ne. # of model nx")
    end if
    if (ox .gt. mdl%xm%o ) then
      write(0,*) ox, mdl%xm%o
      call seperr("Min.x-image space .gt. Min.x-model")
    end if
    ! ----------------------------------------------------------------------------------------------------------------------
    ! The final 2-D grid for image 
	call from_param("mdp_ox",img%xm%o,ox)
	call from_param("mdp_nx",img%xm%n,nx)
	call from_param("mdp_dx",img%xm%d,dx)
	write(0,*) "img%x%n=", img%xm%o 
	write(0,*) "img%x%n=", img%xm%n 
	write(0,*) "img%x%n=", img%xm%d 

	call from_param("mdp_oz",img%z%o,oz)
	call from_param("mdp_nz",img%z%n,nz)
	call from_param("mdp_dz",img%z%d,dz)
	write(0,*) "img%z%n=", img%z%o 
	write(0,*) "img%z%n=", img%z%n 
	write(0,*) "img%z%n=", img%z%d 

    if (img%xm%o .lt. mdl%xm%o ) then
      write(0,*) img%xm%o, mdl%xm%o
      call seperr("Min.x-image space .lt. Min.x-model")
    end if 
    if (img%xm%o + (img%xm%n-1) * img%xm%d .gt. mdl%xm%o + (mdl%xm%n-1)*mdl%xm%d ) then
!      write(0,*) (img%xm%o + (img%xm%n-1) * img%xm%d, mdl%xm%o + (mdl%xm%n-1)*mdl%xm%d 
      call seperr("Max.x-image space .gt. Max.x-model")
    end if 
    if (img%z%o .lt. mdl%z%o ) then
      write(0,*) img%z%o, mdl%z%o
      call seperr("Min.z-image space .lt. Min.z-model")
    end if 
    if (img%z%o + (img%z%n-1) * img%z%d .gt. mdl%z%o + (mdl%z%n-1)*mdl%z%d ) then
!      write(0,*) (img%z%o + (img%z%n-1) * img%z%d, mdl%z%o + (mdl%z%n-1)*mdl%z%d 
      call seperr("Max.z-image space .gt. Max.z-model")
    end if 

    !--------------------------------------------------
    ! Subsurface offset vector for the image
    !
    call from_param("nh",nsecdim,1)
    call from_param("dh",dsecdim,img%xm%d)
    call from_param("oh",osecdim,-int(nsecdim/2)*dsecdim)
    call from_param("mdso", mdso, 1)
    call from_param("alpha", alpha, 0.)
	write(0,*) "mdso", mdso
    write(0,*) "s_offset", "nh=", nsecdim, "dh=", dsecdim, "oh=", osecdim

    ! Read in the mask for seabed 
    allocate(mute(nx*nz*npar_inv))
    call sreed("mask", mute, 4*nx*nz*npar_inv)
	write(0,*) "sum of mute", minval(mute), maxval(mute), npar_inv

   !--------------------------------------------------
    ! Spatial Frequency vector
    !
!    kxm%n = rec%h%n+2*(tapertr+padtraces)
!    kxm%d = 1./(kxm%n*rec%h%d)
!    kxm%o = 1./(2*rec%h%d)-kxm%d
	kxm%n = nx
	kxm%d = 1./(nx*dx)
	kxm%o = 1./(2*dx) - kxm%d
	write(0,*) "spatial frenquency parameters"
	write(0,*) "nkx = ", kxm%n
	write(0,*) "dkx = ", kxm%d
	write(0,*) "okx = ", kxm%o
    allocate(kxm%all(kxm%n)); kxm%all=0.
    do i=1,kxm%n/2+1
      kxm%all(i) = (i-1) * kxm%d
    end do
    do i=kxm%n/2+2, kxm%n
      kxm%all(i) = (i-kxm%n-1) * kxm%d
    end do
    kxm%all = kxm%all * pi*2

    !--------------------------------------------------
    ! Frequency vector
    !
    rec%w%o   = 2*pi * rec%w%o 
    rec%w%d   = 2*pi * rec%w%d  
    allocate(rec%w%all(rec%w%n))
    do iw = 1,rec%w%n
      rec%w%all(iw) = rec%w%o + (iw - 1) * rec%w%d
    end do

    !--------------------------------------------------
    ! Time axis
    ! FFt from frequency
    nt = rec%w%n * 2
    dt = 1./(rec%w%n*nt) * pi*2
    ot = 0.


    !--------------------------------------------------
    ! Compute the IO size  
	mdl%iosize = 4*nx*npar_in*nz 
	img%iosize = 4*img%xm%n*nsecdim*img%z%n 
	rec%iosize = 8*rec%h%n*rec%w%n 
	sou%iosize = 8*sou%w%n 
	gradiosize = 4*nx*npar_inv*nz 
	

    !--------------------------------------------------
    ! Set up the smoothing scheme 
	call from_param("grdsm_style", grdsm_style, "ster")    ! Default is spln 

	select case (grdsm_style) 
	case("spln")
		write(0,*) "Smoothing scheme for velocity grd: SPLN"
		call from_param("axkn_v", axk_v%n, 10); call from_param("axkd_v", axk_v%d, 50.); call from_param("axko_v", axk_v%o, 0.)
		call from_param("azkn_v", azk_v%n, 20); call from_param("azkd_v", azk_v%d, 8. ); call from_param("azko_v", azk_v%o, 0.)
		call from_param("axkn_e", axk_e%n, 10); call from_param("axkd_e", axk_e%d, 50.); call from_param("axko_e", axk_e%o, 0.)
		call from_param("azkn_e", azk_e%n, 20); call from_param("azkd_e", azk_e%d, 8. ); call from_param("azko_e", azk_e%o, 0.)
	case("tria")	
		write(0,*) "Smoothing scheme for velocity grd: TRIA"
		call from_param("rect1_v", rect1_v, 10)
		call from_param("rect2_v", rect2_v, 10)
		call from_param("nsmth_v", nsmth_v, 1 )
		call from_param("rect1_e", rect1_e, 20)
		call from_param("rect2_e", rect2_e, 20)
		call from_param("nsmth_e", nsmth_e, 1 )
	case("ster")	
		write(0,*) "Smoothing scheme for velocity grd: STER"
		if ( .not. exist_file("dip")) then
			write(0,*) "Steering smoothing requires a dip field the same size as model field."
			write(0,*) "Dip field missing, exit"
			stop
		else
			! Dimension of dip /nz, nx/ is different from velocity /nx, nz/	
			call from_aux("dip","n1", n1); call from_aux("dip", "n2", n2);
			if (n1.ne.nz .or. n2.ne.nx) then
				write(0,*) "Dip field grid doesn't match model field grid"	
				write(0,*) "MODEL Dimension: nx=", nx, "nz=", nz
				stop
			else
				allocate(dip(nz*nx))	
				call sreed("dip", dip, 4*nz*nx)
				where(dip < -1.0) dip=-1.0
				where(dip >  1.0) dip= 1.0
				! Steering parameters
				call from_param("nang",nang, 360)
				call from_param("eps_steer",eps_steer,0.01)
				call from_param("nfilt",nfilt,10)
				call from_param("lscale",lscale,0)
				call from_param("destory",destroy,0)
				! smoothing parameters
				call from_param("rect1_v", rect1_v, 10)
				call from_param("rect2_v", rect2_v, 10)
				call from_param("nsmth_v", nsmth_v, 1 )
				call from_param("rect1_e", rect1_e, 20)
				call from_param("rect2_e", rect2_e, 20)
				call from_param("nsmth_e", nsmth_e, 1 )
			end if	
		end if
	case default
		write(0,*) "EXIT: grdsm_style NEEDED"
		write(0,*) "grdsm_style can be"
		write(0,*) "spln for b-spline smoothing"
		write(0,*) "tria for triangle smoothing"
		write(0,*) "ster for steering smoothing"
		stop
	end select	

    ! -------------------------------------------------
    ! works for symmetric offset

	if (exist_file ("grad")) then
		call auxout("grad");
    	call to_history("n1",nx,"grad") ; call to_history("d1",dx,"grad") ; call to_history("o1",ox,"grad")
	    call to_history("n2",nz,"grad") ; call to_history("d2",dz,"grad") ; call to_history("o2",oz,"grad")
   		call to_history("n3",npar_inv,"grad") ; call to_history("d3",1,"grad" ) ; call to_history("o3",0 ,"grad")
   		call to_history("label2", "z(m)","grad"); call to_history("label1", "x(m)","grad"); 
    	call to_history("label3", "par","grad"); call to_history("label4", "iteration");
    	call to_history("esize",4,"grad")
	end if

	if (exist_file("grad_ns")) then
		call auxout("grad_ns");
    	call to_history("n1",nx,"grad_ns") ; call to_history("d1",dx,"grad_ns") ; call to_history("o1",ox,"grad_ns")
    	call to_history("n2",nz,"grad_ns") ; call to_history("d2",dz,"grad_ns") ; call to_history("o2",oz,"grad_ns")
    	call to_history("n3",npar_inv,"grad_ns") ; call to_history("d3",1,"grad_ns" ) ; call to_history("o3",0 ,"grad_ns")
    	call to_history("label2", "z(m)","grad_ns"); call to_history("label1", "x(m)","grad_ns"); 
    	call to_history("label3", "par","grad_ns"); call to_history("label4", "iteration");
    	call to_history("esize",4,"grad_ns")
	end if

	if (exist_file("dmodel")) then
		call auxout("dmodel");
    	call to_history("n1",nx,"dmodel") ; call to_history("d1",dx,"dmodel") ; call to_history("o1",ox,"dmodel")
    	call to_history("n2",nz,"dmodel") ; call to_history("d2",dz,"dmodel") ; call to_history("o2",oz,"dmodel")
    	call to_history("n3",npar_inv,"dmodel") ; call to_history("d3",1,"dmodel" ) ; call to_history("o3",0 ,"dmodel")
    	call to_history("label2", "z(m)","dmodel"); call to_history("label1", "x(m)","dmodel"); 
    	call to_history("label3", "par","dmodel"); call to_history("label4", "iteration");
    	call to_history("esize",4,"dmodel")
	end if

	if (exist_file("illum")) then
    call auxout("illum")
    call to_history("n1",img%xm%n,"illum") ; call to_history("d1",img%xm%d,"illum") ; call to_history("o1",img%xm%o,"illum")
    call to_history("n2",img%z%n,"illum") ; call to_history("d2",img%z%d,"illum") ; call to_history("o2",img%z%o,"illum")
    call to_history("label2", "z(m)","illum"); call to_history("label1", "x(m)","illum"); 
    call to_history("esize",4,"illum")
	end if

	if (exist_file("bgimage")) then
    call auxinout("bgimage"); 
    call to_history("n1",img%xm%n,"bgimage") ; call to_history("d1",img%xm%d,"bgimage") ; call to_history("o1",img%xm%o,"bgimage")
    call to_history("n2",nsecdim,"bgimage") ;    call to_history("d2",dsecdim,"bgimage" ) ;   call to_history("o2",osecdim,"bgimage")
    call to_history("n3",img%z%n,"bgimage") ; call to_history("d3",img%z%d,"bgimage") ; call to_history("o3",img%z%o,"bgimage")
    call to_history("d4",1 ,"bgimage") ;         call to_history("o4",0 ,"bgimage")
    call to_history("label3", "z(m)","bgimage"); call to_history("label1", "x(m)","bgimage");
    call to_history("label2", "soffset","bgimage"); call to_history("label4", "iteration");
    call to_history("esize",4,"bgimage")
	end if

	if (exist_file("dimage")) then
    call to_history("n1",img%xm%n,"dimage") ; call to_history("d1",img%xm%d,"dimage") ; call to_history("o1",img%xm%o,"dimage")
    call to_history("n2",nsecdim,"dimage") ;    call to_history("d2",dsecdim,"dimage" ) ;   call to_history("o2",osecdim,"dimage")
    call to_history("n3",img%z%n,"dimage") ; call to_history("d3",img%z%d,"dimage") ; call to_history("o3",img%z%o,"dimage")
    call to_history("d4",1 ,"dimage") ;         call to_history("o4",0 ,"dimage")
    call to_history("label3", "z(m)","dimage"); call to_history("label1", "x(m)","dimage");
    call to_history("label2", "soffset","dimage"); call to_history("label4", "iteration");
    call to_history("esize",4,"dimage")
	end if

    call to_history("n1",nx) ;   call to_history("d1",dx) ; call to_history("o1",ox)
    call to_history("n2",nz) ;   call to_history("d2",dz) ; call to_history("o2",oz)
    call to_history("n3",npar_in) ; call to_history("d3",1 ) ; call to_history("o3",0 )
    call to_history("label2", "z(m)"); call to_history("label1", "x(m)");
    call to_history("label3", "par"); call to_history("label4", "iteration"); 
    call to_history("esize",4);
  end subroutine aniparam
end module ani_parms_cg
