module ani_phsscr

  use ani_parms_cg
  use eta_pspi

  implicit none

  contains
  !
  ! Anisotropic phase screen
  !
  subroutine aniphsscr(iwa,inode,sign,ddws_s,ddws_e,wv,tflds,model,refr,gat)

    character*3             :: gat
    complex                 :: ddws_s(:), ddws_e(:), wv(:), tflds(:)
    real                    :: dx2, w2, s2, w4, s4, dd, dd2, dd4
    real                    :: model(:,:), refr(:,:)
    complex, allocatable    :: cube_s(:,:), cube_e(:,:)
    integer                 :: sign,ix,iwa
    complex                 :: iwdz
    integer                 :: inode

    real                                        :: s, eta, dlt
    integer                                     :: nref, iref

    allocate(cube_s(nx,nref_max))
    allocate(cube_e(nx,nref_max))

    do iref=1, nref_max
      cube_s(:,iref) = wv
      cube_e(:,iref) = wv
    end do

    call select_ref(refr(:,1), nref)

    do iref = 1, nref
      s   = refr(iref,1) 
      eta = refr(iref,2)
      dlt = refr(iref,3)
      call aniphsscr_exp(iwa, inode, cube_s(:,iref), cube_e(:,iref), tflds,  s, eta, dlt, sign, gat)
    end do

    call interp_lloyd(cube_s, ddws_s, model, refr, scales, nref)
    call interp_lloyd(cube_e, ddws_e, model, refr, scales, nref)

    deallocate(cube_s, cube_e)
  end subroutine aniphsscr

  ! =================================================================================
  ! One step in depth phase screen 
  subroutine aniphsscr_exp(iwa, inode, DATws, DATwe, tflds, s, eta, dlt, sign, gat)
    implicit none
    complex                                     :: DATws(:), DATwe(:), tflds(:)
    real                                        :: dw, dkx, w, dds, dde
    real                                        :: wov, wov2, wov4, kx, kx2, kx4
    integer                                     :: ikx, jkx, iwa, ix, nkx, sign, inode
    character*3					:: gat
    integer*8, allocatable                      :: planF(:)
    integer*8, allocatable                      :: planI(:)

    real     :: s, eta, dlt, a, b
    real     :: phase, time
    complex  :: i

    i = cmplx(0,1)

    allocate(planF(node), planI(node))
    nkx = kxm%n
    dkx = kxm%d * 2*pi

    if (gat.eq."sou") then
      tflds=DATws
      call ani_fftexes(tflds,inode)
      DATws=tflds/sqrt(1.*kxm%n)
      tflds=DATwe
      call ani_fftexes(tflds,inode)
      DATwe=tflds/sqrt(1.*kxm%n)
    else
      tflds=DATws
      call ani_fftexer(tflds,inode)
      DATws=tflds/sqrt(1.*kxm%n)
      tflds=DATwe
      call ani_fftexer(tflds,inode)
      DATwe=tflds/sqrt(1.*kxm%n)
    end if

    ! anisotropic parameters
    a = 0.5*(1.0+2.0*dlt)
    b = 2.0*eta*(1.0+2.0*dlt) + 0.25*(1.0+2.0*dlt)

    w = rec%w%all(iwa)
    if(w.eq.0.0) w=1.0e-06
    wov  = w    * s
    wov2 = wov  * wov
    wov4 = wov2 * wov2

    do ikx = 1, nkx
      kx = kxm%all(ikx)
      kx2 = kx  * kx
      kx4 = kx2 * kx2
      dds = wov * ( 1 + (a*kx2) / (wov2) + (3*a*b*kx4) / (wov4) ) 
      dde = - (2*a*kx4) / (wov2*wov)
      DATws(ikx) = DATws(ikx) * dds
      DATwe(ikx) = DATwe(ikx) * dde
    end do

    if (gat.eq."sou") then
      tflds=DATws
      call ani_ifftexes(tflds,inode)
      DATws=tflds/sqrt(1.*kxm%n)
      tflds=DATwe
      call ani_ifftexes(tflds,inode)
      DATwe=tflds/sqrt(1.*kxm%n)
    else
      tflds=DATws
      call ani_ifftexer(tflds,inode)
      DATws=tflds/sqrt(1.*kxm%n)
      tflds=DATwe
      call ani_ifftexer(tflds,inode)
      DATwe=tflds/sqrt(1.*kxm%n)
    end if

    DATws = DATws * i * sign * dz
    DATwe = DATwe * i * sign * dz
    deallocate(planF, planI) 
  end subroutine aniphsscr_exp
end module
