module ani_smthgrd

  use sep
  use ani_parms_cg
  use base_spline_mod
  use base_trismt_mod
  use killwave_ns 

  implicit none
  integer, private							:: curiter

contains

  subroutine smoothgrd_init(iter_in)
    integer, intent(in)						:: iter_in
	curiter = iter_in
  end subroutine

  subroutine smoothgrd_spln(model)
    real, allocatable                       :: temp_v(:), temp_e(:)
    real, allocatable                       :: gspl_v(:), gspl_e(:)
	real, intent(inout)						:: model(:)
	integer									:: stat, stat1
	write(0,*) "aaaaaa"

    ! Scale and Smooth the gradient field
    allocate(temp_v(nx*nz), temp_e(nx*nz))
    allocate(gspl_v(axk_v%n*azk_v%n), gspl_e(axk_e%n*azk_e%n))
    temp_v = model(1:nx*nz);
    temp_e = model(nx*nz+1:);
    gspl_v = 0.; gspl_e = 0.;

    ! Smooth velocity
    stat = base_spln2d_init(mdl%z, mdl%xm, azk_v, axk_v)
    stat1 = base_spln2d_oper(.true. , .false., gspl_v, temp_v)     
    stat1 = base_spln2d_oper(.false., .false., gspl_v, temp_v)

    ! Smooth eta 
    stat = base_spln2d_init(mdl%z, mdl%xm, azk_e, axk_e)
    stat1 = base_spln2d_oper(.true. , .false., gspl_e, temp_e)     
    stat1 = base_spln2d_oper(.false., .false., gspl_e, temp_e)

    ! Scale the gradient field according to the iteration #
    ! For the first iter_v iterations, update v only
    if (curiter <= iter_v) then
      temp_v = temp_v * 1e-20 * 1e-4 * g_scales(1) 
      temp_e = temp_e * 1e-20 * 1e-4 * g_scales(2) 
    else
      temp_v = temp_v * 1e-20 * 1e-4 * g_scales(1) 
      temp_e = temp_e * 1e-20 * 1e-4 * g_scales(2) 
    end if

    model(1:nx*nz) = temp_v;
    model(nx*nz+1:) = temp_e;

	deallocate(temp_v,temp_e)
	deallocate(gspl_v,gspl_e)

  end subroutine smoothgrd_spln 

  subroutine smoothgrd_tria(model)
    real, allocatable                       :: temp_v(:,:), temp_e(:,:)
	real, intent(inout)						:: model(:)
	integer									:: stat	

	write(0,*) "aaaaaa"
	! Scale and Smooth the gradient field
    allocate(temp_v(nx,nz), temp_e(nx,nz))
    temp_v = reshape(model(1:nx*nz), (/nx,nz/));
    temp_e = reshape(model(nx*nz+1:),(/nx,nz/));

	stat = base_trismt_init(nx,nz,1)
	! Smooth velocity
	stat = base_tris2d_oper(temp_v, rect1_v, rect2_v, nsmth_v, .true. )
	stat = base_tris2d_oper(temp_v, rect1_v, rect2_v, nsmth_v, .false.)
	! Smooth eta	
	stat = base_tris2d_oper(temp_e, rect1_e, rect2_e, nsmth_e, .true. )
	stat = base_tris2d_oper(temp_e, rect1_e, rect2_e, nsmth_e, .false.)

    ! Scale the gradient field according to the iteration #
    ! For the first iter_v iterations, update v only
    if (curiter <= iter_v) then
      temp_v = temp_v / maxval(abs(temp_v)) * g_scales(1) 
      temp_e = temp_e / maxval(abs(temp_e)) * g_scales(2) 
    else
      temp_v = temp_v / maxval(abs(temp_v)) * g_scales(1) 
      temp_e = temp_e / maxval(abs(temp_e)) * g_scales(2) 
    end if

    model(1:nx*nz) = reshape(temp_v, (/nx*nz/));
    model(nx*nz+1:) = reshape(temp_e, (/nx*nz/));

	deallocate(temp_v,temp_e)

  end subroutine smoothgrd_tria 

  subroutine smoothgrd_ster(model)
    real, allocatable                       :: temp_model(:), temp_v(:), temp_e(:)
	real, intent(inout)						:: model(:)
	integer									:: nfilt
	real 									:: oang,dang

    oang = minval(dip)
    dang = (maxval(dip)-minval(dip)) / (nang-1)
    write(0,*) "min ang", oang
    write(0,*) "max ang", maxval(dip), oang + (nang-1)*dang
	
	allocate(temp_model(nz*nx),temp_v(nz*nx),temp_e(nz*nx))	
	call ns_init(nz,nx,1,dang,oang)
	temp_v = reshape(transpose(reshape(model(1:nx*nz ),(/nx,nz/))), (/nz*nx/) )
	call ns_lop(temp_v, temp_model, .false.)
	temp_v = reshape(transpose(reshape(temp_model,(/nz,nx/))), (/nx*nz/))
	model(1:nx*nz)  = temp_v  / maxval(abs(temp_v)) * g_scales(1)

	temp_e = reshape(transpose(reshape(model(nx*nz+1:),(/nx,nz/))), (/nz*nx/) )
	call ns_lop(temp_e, temp_model, .false.)
	temp_e = reshape(transpose(reshape(temp_model,(/nz,nx/))), (/nx*nz/))
	model(nx*nz+1:)  = temp_e / maxval(abs(temp_e)) * g_scales(2)

	deallocate(temp_model, temp_e, temp_v)
	call ns_release()

  end subroutine smoothgrd_ster 

end module ani_smthgrd
