module ani_steepds_rand

  use sep
  use ani_type
  use ani_parms_cg
  use ani_exec_time
  use ani_taper
  use ani_wemvacg2
  use igrad1
  use base_spline_mod
  use random_mod
  use ani_smthgrd

  implicit none

  integer, private                                       :: iter, nreal, seekpos
  complex, allocatable, private                          :: source1(:)
  real, allocatable, private                             :: taper(:)

  contains

  !======================================================================
  subroutine steepds_init(src_in, taper_in)

    complex                             :: src_in(:)
    real                                :: taper_in(:)

    if (allocated(source1)) deallocate(source1)
    if (allocated(taper))  deallocate(taper)

    allocate(source1(sou%w%n), taper(nx))

    source1 = src_in
    taper  = taper_in

  end subroutine steepds_init
  !======================================================================

  !======================================================================
  subroutine fgeval_init(iter_in)

    integer                             :: iter_in
    iter = iter_in
	if (ec_style .eq. 0) then
		nreal = 30
	else 	
		nreal = iter*10
	end if	

  end subroutine fgeval_init
  !======================================================================

  !======================================================================
  subroutine feval(model_in, refr_in, Bimage)

    real, intent(inout)                                 :: model_in(:,:,:), refr_in(:,:,:)
    real, intent(out)                                   :: Bimage(:,:,:)

    real, allocatable                                   :: image(:,:,:)
    complex, allocatable								:: recv(:,:),source(:,:)
    complex, allocatable                                :: recv_temp(:,:)
    integer												:: ish, sh_position, iz, ix, isecdim, ipar, stat, stat1, rec_pos
    real                                                :: subh, start, finish, depth
	integer												:: ireal
	real, allocatable									:: rand_code(:)
	integer, external									:: sreed, srite, sseek

    allocate(source(nx, sou%w%n)); source = 0.
    allocate(recv(nx, rec%w%n))  ; recv   = 0.
    allocate(image(img%xm%n,nsecdim,img%z%n))
	if(.not. allocated(rand_code)) allocate(rand_code(rec%xd%n))
    if(.not. allocated(recv_temp)) allocate(recv_temp(rec%h%n, rec%w%n)); recv_temp   = 0.;

    call aniwei_init(iter)

    start=0; finish=0; 
    call aniexec_time(start)

    Bimage=0.; 

	if (ifec_img) then
			write(0,*) "nreal = ", nreal
		do ireal = 1, nreal
			source=0.; recv=0; 
				! Setup the random schedule
			call random_init((iter-1)*nreal+ireal)
			call rand_real_array_1d(rand_code)
			! Turn to random -1 and 1 
			call sign_demean(rand_code)
			seekpos = sseek("rec",0,0)	

			do ish = 1, nsh
				recv_temp = 0.
				sh_position = ((ish-1)*dsh + osh - ox) / dx + 1
				source(sh_position,:) =source(sh_position,:) + rand_code(ish) * conjg(source1);
				if (rec%iosize .ne. sreed("rec", recv_temp, rec%iosize)) then
						call seperr("Input size of rec .ne. size of rec")
				end if		
				rec_pos = ((ish-1)*dsh + osh + rec%h%o -ox ) /dx + 1
				recv(rec_pos:rec_pos+rec%h%n-1,:) = recv(rec_pos:rec_pos+rec%h%n-1,:)  + rand_code(ish) * conjg(recv_temp)
			end do
			image=0; call aniwei(taper,recv,source,model_in,refr_in,image)
			Bimage= Bimage + image
		end do
		Bimage = Bimage / nreal
    else
		write(0,*) "start imaging"	
		seekpos = sseek("rec",0,0)
		do ish = 1, nsh
		    recv=0.; recv_temp=0. 
			source = 0.;
			sh_position = ((ish-1)*dsh + osh - ox) / dx + 1
!			write(0,*) "shot pos", sh_position
			source(sh_position,:) = conjg(source1);
			if (rec%iosize .ne. sreed("rec", recv_temp, rec%iosize)) then
					call seperr("Input size of rec .ne. size of rec")
			end if		
			rec_pos = ((ish-1)*dsh + osh + rec%h%o -ox ) /dx + 1
!			write(0,*) "rec_pos", rec_pos
			recv(rec_pos:rec_pos+rec%h%n-1,:) = conjg(recv_temp)
            image=0; call aniwei(taper,recv,source,model_in,refr_in,image)
            Bimage= Bimage + image
			if (ish/10==0) write(0,*) "after imaging", ish, "of", nsh, "shots"
        end do
    end if

    call aniexec_time(finish)
    write(0,112) "COMPUTED BACKGROUND IMAGE ", (finish-start)/60," min"         

    deallocate(image,source,recv)
	if (allocated(recv_temp)) deallocate(recv_temp) 
	if (allocated(rand_code)) deallocate(rand_code)		
    112 format(a29,f10.2,a20,i4)
  end subroutine feval

  !======================================================================
  !======================================================================

  subroutine geval(model_in, refr_in, Bimage, grad)

    real,intent(in)                                     :: model_in(:,:,:), refr_in(:,:,:)
    real,intent(inout)                                  :: Bimage(:,:,:)
    real,intent(out)                                    :: grad(:)

    integer						:: ish, sh_position, iz, ix, isecdim, ipar, stat, stat1
    real                                                :: start, finish
    real, allocatable                                   :: dmodel_in(:)
	integer, external									:: sreed, srite


    ! Compute the gradient using the adjoint wemva
    start=0; finish=0; 
    call aniexec_time(start)
	call aniwei_init(iter)
    call aniwemva_init(source1, taper, model_in, refr_in)
    allocate(dmodel_in(nz*nx*npar_inv)); dmodel_in=0.
	if (ifec_grd) then
			call aniwemva_lop_rand(.true., .false., dmodel_in, Bimage)
    else
			call aniwemva_lop2(.true., .false., dmodel_in, Bimage)
    end if

	! Output non-smooth gradient	
	if ( gradiosize .ne. srite("grad_ns", dmodel_in, gradiosize) ) then
			call seperr("Output size of grad .ne. size of grad")
	end if	
    call to_history("n4", iter, "grad_ns")


	! Initialize the smoother
	call smoothgrd_init(iter)

	select case (grdsm_style)
	case("spln")
		dmodel_in = dmodel_in * mute 
		call smoothgrd_spln(dmodel_in)
	case("tria")
		dmodel_in = dmodel_in * mute 
		call smoothgrd_tria(dmodel_in)
	case("ster")
		call smoothgrd_ster(dmodel_in)
		dmodel_in = dmodel_in * mute 
		call smoothgrd_tria(dmodel_in)
	case default
		write(0,*) "EXIT: grdsm_style NEEDED"
		write(0,*) "grdsm_style can be"
		write(0,*) "spln for b-spline smoothing"
		write(0,*) "tria for triangle smoothing"
		write(0,*) "ster for steering smoothing"
		stop
	end select

    ! Mute the dmodel down to seabed
    grad = -dmodel_in

	! Output smoothed and muted gradient	
	if ( gradiosize .ne. srite("grad", dmodel_in, 4*nz*nx*npar_inv) ) then
			call seperr("Output size of grad .ne. size of grad")
	end if	
    call to_history("n4", iter, "grad")

    call aniexec_time(finish)
    write(0,112) "COMPUTED MODEL GRADIENT ",(finish-start)/60," min "
    if (allocated(dmodel_in)) deallocate(dmodel_in)

	call aniwemva_close()		

    111 format(a26,i4,a8,i4,a4,f6.2,a4)
    112 format(a29,f10.2,a20,i4)

  end subroutine geval
  !======================================================================

end module ani_steepds_rand
