module ani_wemvacg2

  use sep
  use ani_type
  use ani_fftw
  use ani_parms_cg
  use ani_image_deci
  use eta_pspi
  use ani_phsscr
  use adj_mod
  use ani_exec_time
  use random_mod

  implicit none

  integer, private                                       :: iter, nodes, nreal
  complex, allocatable, private                          :: source1(:)
  real, allocatable, private                             :: taper(:), model(:,:,:), refr(:,:,:)

  contains

  !======================================================================
  subroutine sign_demean(input)
    real, intent(inout)					:: input(:)
	real, allocatable					:: temp(:)
	
	allocate(temp(size(input)))
	temp = input -0.5;
	where (temp .lt. 0) input = -1.;
	where (temp .ge. 0) input =  1.;
	deallocate(temp)
  end subroutine

  !======================================================================
  subroutine aniwei_init(iter_in)
    integer, external			:: omp_get_thread_num,omp_get_num_threads
	integer						:: iter_in
    if (node.eq.9999) then
      !$OMP PARALLEL
      nodes = omp_get_num_threads()
      !$OMP END PARALLEL
    else
      nodes=node
    end if
    write(3,*) "****************"
    write(3,*) "* NUM.THREADS = ",nodes
    write(3,*) "****************"
	iter = iter_in
	if (ec_style .eq. 0) then
		nreal = 20
	else 	
		nreal = iter*10
	end if	

  end subroutine aniwei_init

  !======================================================================
  subroutine aniwei(taper,recv,source,model_in,refr_in,image)

    complex						:: recv(:,:),source(:,:)
    integer						:: iwa,iz,inode,i,ixw,isecdim
    real						:: model_in(:,:,:), refr_in(:,:,:)
    real						:: image(:,:,:), taper(:)
    real, allocatable					:: GGG(:,:,:,:) 
    real, allocatable					:: model(:,:,:), refr(:,:,:) 
    complex, allocatable				:: ws(:,:),wr(:,:)
    complex, allocatable				:: tflds(:,:),tfldr(:,:)
    integer, external   			        :: omp_get_thread_num,omp_get_num_threads

    allocate(ws(nx,nodes), wr(nx,nodes)); wr=0.; ws=0.
    allocate(tflds(nx,nodes), tfldr(nx,nodes)); tflds=0.;tfldr=0.
    allocate(GGG(img%xm%n,nsecdim,img%z%n,nodes)); GGG=0.
    allocate(model(nx, nz, npar_in), refr(nref_max, nz, npar_in)); 
    model = model_in; refr  = refr_in

    ! Setup the fftw plans 
    call ani_fftws(tflds, nodes)
    call ani_fftwr(tfldr, nodes)

    !$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)
    do iwa = 1, rec%w%n
      iz = 1; inode=omp_get_thread_num()+1
      ws(:,inode)=0.;wr(:,inode)=0.
      ws(tapertr+padtraces+1:nx-padtraces-tapertr-1,inode) = source(:,iwa)
      wr(tapertr+padtraces+1:nx-padtraces-tapertr-1,inode) =   recv(:,iwa)
      call aniimage(.false., iwa, ws(:,inode), wr(:,inode), GGG(:,:,iz,inode))   
      do iz = 2, nz
        call etapspi(iwa, inode, -1, ws(:,inode), tflds(:,inode), model(:,iz-1,:), refr(:,iz-1,:), "sou")
        call etapspi(iwa, inode,  1, wr(:,inode), tfldr(:,inode), model(:,iz-1,:), refr(:,iz-1,:), "rec")
        ws(:,inode)=ws(:,inode)*taper
        wr(:,inode)=wr(:,inode)*taper
        call aniimage(.false., iwa, ws(:,inode), wr(:,inode), GGG(:,:,iz,inode))
      end do
    end do
    !$OMP END PARALLEL DO

    do inode=1,nodes
      !$OMP parallel do private(ixw,iz,isecdim)
      do iz=1,img%z%n
        do isecdim=1,nsecdim
          do ixw=1,img%xm%n
            image(ixw,isecdim,iz)=image(ixw,isecdim,iz)+GGG(ixw,isecdim,iz,inode)
          end do
        end do
      end do
      !$OMP END PARALLEL DO
    end do
    6   format(A13,i3,A8,i3,A14,i3,A12,i3)
    deallocate(ws, wr, GGG, model, refr, tflds, tfldr)
  end subroutine aniwei




  !======================================================================
  subroutine aniwemva_init(src_in, taper_in, model_in, refr_in)

    integer                             :: idip_in, nodes
    integer, external			:: omp_get_thread_num,omp_get_num_threads
    complex                             :: src_in(:)
    real                                :: taper_in(:), model_in(:,:,:), refr_in(:,:,:)
    if (node.eq.9999) then
      !$OMP PARALLEL
      nodes = omp_get_num_threads()
      !$OMP END PARALLEL
    else
      nodes=node
    end if

    write(3,*) "****************"
    write(3,*) "* NUM.THREADS = ",nodes
    write(3,*) "****************"

    if (allocated(source1)) deallocate(source1)
    if (allocated(taper))  deallocate(taper)
    if (allocated(model))  deallocate(model)
    if (allocated(refr))   deallocate(refr)

    allocate(source1(sou%w%n), taper(nx))
    allocate(model(nx,nz,npar_in), refr(nref_max,nz,npar_in))

    source1 = src_in
    taper  = taper_in
    model  = model_in
    refr   = refr_in
  
  end subroutine aniwemva_init
  !======================================================================

  !======================================================================
  function aniwemva_lop(adj, add, modl, data) result(stat)

    integer                :: stat
    logical, intent(in)    :: adj, add
    real, dimension(:)     :: modl, data

    stat = 1
    call adjnull(adj, add, modl, data)
    call aniwemva_lop2(adj, add, modl, data)
    stat = 0

  end function aniwemva_lop
  !======================================================================

  !======================================================================
  subroutine aniwemva_lop2(adj, add, deltaM, deltaI)
    logical						:: adj, add
    integer						:: iwa,iz,inode,isecdim,ixw,ish
    real, intent(inout)					:: deltaM(nx,nz,npar_inv)
    real, intent(inout)       				:: deltaI(img%xm%n,nsecdim,img%z%n)
    real, allocatable					:: SSS(:,:,:), EEE(:,:,:)
    real, allocatable					:: GGG(:,:,:,:)
    complex, allocatable				:: ws(:,:),wr(:,:)
    complex, allocatable				:: tflds(:,:),tfldr(:,:)
    complex, allocatable                                :: source(:,:), recv(:,:)
    complex, allocatable                                :: recv_temp(:,:)
    complex, allocatable				:: ddws_s(:,:),ddwr_s(:,:)
    complex, allocatable				:: ddws_e(:,:),ddwr_e(:,:)
    complex, allocatable				:: dws(:,:,:),dwr(:,:,:)
    complex, allocatable				:: bws(:,:,:),bwr(:,:,:)
    integer, external   			        :: omp_get_thread_num,omp_get_num_threads
    real                                                :: start, finish, start1, finish1
    integer                                             :: sh_position, rec_pos, seek_pos
	integer, external									:: sreed, srite, sseek

    allocate(bws(nx,nz,nodes), bwr(nx,nz,nodes)); bwr=0; bws=0
    allocate(ws(nx,nodes), wr(nx,nodes)); ws=0; wr=0
    allocate(tflds(nx,nodes), tfldr(nx,nodes)); tflds=0.;tfldr=0.
    allocate(ddws_s(nx,nodes), ddwr_s(nx,nodes)); ddws_s=0; ddwr_s=0
    allocate(ddws_e(nx,nodes), ddwr_e(nx,nodes)); ddws_e=0; ddwr_e=0
    if(.not. allocated(source))    allocate(source   (nx,      sou%w%n))
    if(.not. allocated(recv))      allocate(recv     (nx,      rec%w%n))
    if(.not. allocated(recv_temp)) allocate(recv_temp(rec%h%n, rec%w%n)); recv_temp   = 0.;

    if (adj) then
      deltaM=0.
      write(0,*) "Adjoint: zero out deltaM", nodes
    else
      deltaI=0.
      write(0,*) "Forward: zero out deltaI", nodes
    end if

    ! Setup the fftw plans 
    call ani_fftws(tflds, nodes)
    call ani_fftwr(tfldr, nodes)

    seek_pos = sseek("rec",0,0)
    
    ! ----------------------------------------
    ! Shot loop
    do ish = 1, nsh
      start=0; finish=0; source = 0.; recv=0
      call aniexec_time(start)
      sh_position = ((ish-1)*dsh + osh - ox) / dx + 1
!      write(0,*) "shot position=", sh_position, (ish-1)*dsh + osh
      source(sh_position,:) = conjg(source1);
	  if (rec%iosize .ne. sreed("rec", recv_temp, rec%iosize)) then
			call seperr("Input size of rec .ne. size of rec")
	  end if		
      rec_pos = ((ish-1)*dsh + osh + rec%h%o -ox ) /dx + 1
	  recv(rec_pos:rec_pos+rec%h%n-1,:) = conjg(recv_temp)

      ! ----------------------------------------
      ! Adjoint tomographic operator
      if (adj) then
        allocate(SSS(nx,nz,nodes),EEE(nx,nz,nodes))
        SSS=0.                                 ! Temporary location for slowness perturbation
        EEE=0.                                 ! Temporary location for \eta perturbation 
        allocate(dwr(1,kxm%n,nodes));     dwr=0.
        allocate(dws(1,kxm%n,nodes));     dws=0.

        !$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)       
        do iwa = 1, rec%w%n
          ! Initializing 
          inode=omp_get_thread_num()+1; 
          ws(:,inode)=0.;wr(:,inode)=0.
          ws(tapertr+padtraces+1:nx-padtraces-tapertr-1,inode) = source(:,iwa)     ! Source and recv are not padded
          wr(tapertr+padtraces+1:nx-padtraces-tapertr-1,inode) =   recv(:,iwa)
          dws(1,:,inode) = 0.
          dwr(1,:,inode) = 0.
          bws(:,:,inode) = 0.
          bwr(:,:,inode) = 0.
          bws(:,1,inode) = ws(:,inode)
          bwr(:,1,inode) = wr(:,inode)

          ! Background wavefields, bws & bwr
          do iz = 2, img%z%n 
            call etapspi(iwa, inode, -1, ws(:,inode), tflds(:,inode), model(:,iz-1,:), refr(:,iz-1,:),"sou")
            call etapspi(iwa, inode,  1, wr(:,inode), tfldr(:,inode), model(:,iz-1,:), refr(:,iz-1,:),"rec")
            ws(:,inode)=ws(:,inode)*taper
            wr(:,inode)=wr(:,inode)*taper
            bws(:,iz,inode)=ws(:,inode)
            bwr(:,iz,inode)=wr(:,inode)
          end do

          do iz=img%z%n-1, 1, -1
            ! Adjoint imaging condition to get the wavefield perturbation
            call aniimage(.true., iwa, dws(1,:,inode),    bwr(:,iz+1,inode), deltaI(:,:,iz+1),  .true.)
            call aniimage(.true., iwa, bws(:,iz+1,inode), dwr(1,:,inode),    deltaI(:,:,iz+1), .false.)

            ! Upward propagate wavefield perturbations
            call etapspi(iwa, inode, 1, dws(1,:,inode), tflds(:,inode), model(:,iz,:), refr(:,iz,:), "sou")
            dws(1,:,inode)=dws(1,:,inode)*taper
            call etapspi(iwa, inode,-1, dwr(1,:,inode), tfldr(:,inode), model(:,iz,:), refr(:,iz,:), "rec")
            dwr(1,:,inode)=dwr(1,:,inode)*taper

            ! Phase-screen operator & adjoint scattering operator to get the model perturbation
            ddws_s(:,inode)=0; ddws_e(:,inode)=0; 
            call aniphsscr(iwa,inode, 1,ddws_s(:,inode),ddws_e(:,inode),bws(:,iz,inode),tflds(:,inode),model(:,iz,:),refr(:,iz,:),"sou")
            SSS(:,iz,inode)=SSS(:,iz,inode)-conjg(ddws_s(:,inode))*(dws(1,:,inode))/rec%w%n    
            EEE(:,iz,inode)=EEE(:,iz,inode)-conjg(ddws_e(:,inode))*(dws(1,:,inode))/rec%w%n    

            ddwr_s(:,inode)=0; ddwr_e(:,inode)=0;
            call aniphsscr(iwa,inode,-1,ddwr_s(:,inode),ddwr_e(:,inode),bwr(:,iz,inode),tflds(:,inode),model(:,iz,:),refr(:,iz,:),"rec")
            SSS(:,iz,inode)=SSS(:,iz,inode)-conjg(ddwr_s(:,inode))*(dwr(1,:,inode))/rec%w%n    
            EEE(:,iz,inode)=EEE(:,iz,inode)-conjg(ddwr_e(:,inode))*(dwr(1,:,inode))/rec%w%n    
          end do

          !              if (mod(iwa,50)==1) write(3,6) " FREQUENCY # ",iwa," OUT OF ",rec%w%n," / SYNTH.FCT. # ",rec%xd%n, &
          !                   " FROM NODE: ",inode
        end do
        !$OMP END PARALLEL DO

        do inode=1,nodes
          !$OMP parallel do private(ixw,iz)
          do iz=1,nz
            do ixw=1,kxm%n
              deltaM(ixw,iz,1)=deltaM(ixw,iz,1)+SSS(ixw,iz,inode)
              deltaM(ixw,iz,2)=deltaM(ixw,iz,2)+EEE(ixw,iz,inode)
            end do
          end do
          !$OMP END PARALLEL DO
        end do
        deallocate(dwr,dws,EEE,SSS)
        ! ----------------------------------------
        ! Forward tomographic operator
      else 
        allocate(GGG(img%xm%n,nsecdim,img%z%n,nodes))
        GGG=0.
        allocate(dwr(1,kxm%n,nodes));               dwr=0.
        allocate(dws(1,kxm%n,nodes));               dws=0.
        !$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)
        do iwa = 1,rec%w%n
          inode=omp_get_thread_num()+1;
          ws(:,inode)=0.;wr(:,inode)=0.
          ws(tapertr+padtraces+1:nx-padtraces-tapertr-1,inode) = source(:,iwa)
          wr(tapertr+padtraces+1:nx-padtraces-tapertr-1,inode) =   recv(:,iwa)
          dws(1,:,inode)=0.
          dwr(1,:,inode)=0.
          bws(:,:,inode)=0.
          bwr(:,:,inode)=0.
          bws(:,1,inode)=ws(:,inode)
          bwr(:,1,inode)=wr(:,inode)

          do iz = 2, img%z%n 
            call etapspi(iwa, inode, -1, ws(:,inode), tflds(:,inode), model(:,iz-1,:), refr(:,iz-1,:), "sou")
            call etapspi(iwa, inode,  1, wr(:,inode), tfldr(:,inode), model(:,iz-1,:), refr(:,iz-1,:), "rec")
            ws(:,inode)=ws(:,inode)*taper
            wr(:,inode)=wr(:,inode)*taper
            bws(:,iz,inode)=ws(:,inode)
            bwr(:,iz,inode)=wr(:,inode)
          end do

          do iz=1, img%z%n-1
            ! Phase-screen
            ddws_s(:,inode)=0; ddws_e(:,inode)=0; 
            call aniphsscr(iwa,inode,-1,ddws_s(:,inode),ddws_e(:,inode),bws(:,iz,inode),tflds(:,inode),model(:,iz,:),refr(:,iz,:),"sou")
            dws(1,:,inode)=dws(1,:,inode)+(ddws_s(:,inode))*deltaM(:,iz,1)/rec%w%n + (ddws_e(:,inode))*deltaM(:,iz,2)/rec%w%n 
            
            ! Scattering
            call etapspi(iwa, inode, -1, dws(1,:,inode), tflds(:,inode), model(:,iz,:), refr(:,iz,:),"sou")
            dws(1,:,inode)=dws(1,:,inode)*taper

            ! Imaging condition
            call aniimage(.false., iwa, dws(1,:,inode), bwr(:,iz+1,inode), GGG(:,:,iz+1,inode))

            ! Phase-screen
            ddwr_s(:,inode)=0; ddwr_e(:,inode)=0; 
            call aniphsscr(iwa,inode, 1,ddwr_s(:,inode),ddwr_e(:,inode),bwr(:,iz,inode),tflds(:,inode),model(:,iz,:),refr(:,iz,:),"rec")
            dwr(1,:,inode)=dwr(1,:,inode)+(ddwr_s(:,inode))*deltaM(:,iz,1)/rec%w%n + (ddwr_e(:,inode))*deltaM(:,iz,2)/rec%w%n 

            ! Scattering
            call etapspi(iwa, inode,  1, dwr(1,:,inode), tfldr(:,inode), model(:,iz,:), refr(:,iz,:),"rec")
            dwr(1,:,inode)=dwr(1,:,inode)*taper

            ! Imaging condition
            call aniimage(.false., iwa, bws(:,iz+1,inode), dwr(1,:,inode), GGG(:,:,iz+1,inode))
          end do

 !         if (mod(iwa,2)==1) 
!			write(0,6) " FREQUENCY # ",iwa," OUT OF ",rec%w%n," / SYNTH.FCT. # ",rec%xd%n, " FROM NODE: ",inode
        end do
        !$OMP END PARALLEL DO

        do inode=1,nodes
          !$OMP parallel do private(ixw,iz,isecdim)
          do iz=1,img%z%n
            do isecdim=1,nsecdim
              do ixw=1,img%xm%n
                deltaI(ixw,isecdim,iz)=deltaI(ixw,isecdim,iz)+GGG(ixw,isecdim,iz,inode)
              end do
            end do
          end do
          !$OMP END PARALLEL DO
        end do
        deallocate(dwr,dws,GGG)

      end if
    end do
    deallocate(bws,bwr,ws,wr,ddws_s,ddws_e,ddwr_s,ddwr_e,tflds,tfldr)
    deallocate(source, recv, recv_temp)
    6   format(A13,i3,A8,i3,A14,i3,A12,i3)

  end subroutine aniwemva_lop2
  !======================================================================
  !======================================================================
  subroutine aniwemva_lop_rand(adj, add, deltaM, deltaI)
    logical						:: adj, add
    integer						:: iwa,iz,inode,isecdim,ixw,ish
    real, intent(inout)					:: deltaM(nx,nz,npar_inv)
    real, intent(inout)       				:: deltaI(img%xm%n,nsecdim,img%z%n)
    real, allocatable					:: SSS(:,:,:), EEE(:,:,:)
    real, allocatable					:: GGG(:,:,:,:)
    complex, allocatable				:: ws(:,:),wr(:,:)
    complex, allocatable				:: tflds(:,:),tfldr(:,:)
    complex, allocatable                                :: source(:,:), recv(:,:)
    complex, allocatable                                :: recv_temp(:,:)
    complex, allocatable				:: ddws_s(:,:),ddwr_s(:,:)
    complex, allocatable				:: ddws_e(:,:),ddwr_e(:,:)
    complex, allocatable				:: dws(:,:,:),dwr(:,:,:)
    complex, allocatable				:: bws(:,:,:),bwr(:,:,:)
    integer, external   			        :: omp_get_thread_num,omp_get_num_threads
    real                                                :: start, finish, start1, finish1
    integer                                             :: sh_position, ireal, rec_pos, seek_pos
	real, allocatable									:: rand_code(:)
	integer, external									:: sreed, srite, sseek

    allocate(bws(nx,nz,nodes), bwr(nx,nz,nodes)); bwr=0; bws=0
    allocate(ws(nx,nodes), wr(nx,nodes)); ws=0; wr=0
    allocate(tflds(nx,nodes), tfldr(nx,nodes)); tflds=0.;tfldr=0.
    allocate(ddws_s(nx,nodes), ddwr_s(nx,nodes)); ddws_s=0; ddwr_s=0
    allocate(ddws_e(nx,nodes), ddwr_e(nx,nodes)); ddws_e=0; ddwr_e=0
    if(.not. allocated(source)) allocate(source(nx, sou%w%n)); source = 0.;
    if(.not. allocated(recv))   allocate(recv  (nx, rec%w%n)); recv   = 0.;
    if(.not. allocated(recv_temp)) allocate(recv_temp  (rec%h%n, rec%w%n)); recv_temp   = 0.;
	if(.not. allocated(rand_code)) allocate(rand_code(rec%xd%n))

    if (adj) then
      deltaM=0.
      write(0,*) "Adjoint: zero out deltaM", nodes
    else
      deltaI=0.
      write(0,*) "Forward: zero out deltaI", nodes
    end if

    ! Setup the fftw plans 
    call ani_fftws(tflds, nodes)
    call ani_fftwr(tfldr, nodes)
    
    ! Random encoding realization loop
    do ireal = 1, nreal
	
	recv = 0; source = 0;

	! Setup the random schedule
	call random_init((iter-1)*nreal+ireal)
	call rand_real_array_1d(rand_code)
	! Turn to random -1 and 1 
	call sign_demean(rand_code)
	if (any(rand_code==0)) then
		write(0,*) "zero rand_code"
	end if
	seek_pos = sseek("rec",0,0)
	start=0; finish=0; 
	call aniexec_time(start)

	! ----------------------------------------
	! Encode source and receiver
	! ----------------------------------------
	do ish = 1, nsh
    	recv_temp=0
    	sh_position = ((ish-1)*dsh + osh - ox) / dx + 1
!       write(0,*) "shot position=", sh_position, (ish-1)*dsh + osh
    	source(sh_position,:) = source(sh_position,:) + rand_code(ish) * conjg(source1);
		if (rec%iosize .ne. sreed("rec", recv_temp, rec%iosize)) then
				call seperr("Input size of rec .ne. size of rec")
		end if		
		rec_pos = ((ish-1)*dsh + osh + rec%h%o -ox ) /dx + 1
		recv(rec_pos:rec_pos+rec%h%n-1,:) = recv(rec_pos:rec_pos+rec%h%n-1,:)+rand_code(ish)*conjg(recv_temp)
	end do

      ! ----------------------------------------
      ! Adjoint tomographic operator
      if (adj) then
        allocate(SSS(nx,nz,nodes),EEE(nx,nz,nodes))
        SSS=0.                                 ! Temporary location for slowness perturbation
        EEE=0.                                 ! Temporary location for \eta perturbation 
        allocate(dwr(1,kxm%n,nodes));     dwr=0.
        allocate(dws(1,kxm%n,nodes));     dws=0.

        !$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)       
        do iwa = 1, rec%w%n
          ! Initializing 
          inode=omp_get_thread_num()+1; 
          ws(:,inode)=0.;wr(:,inode)=0.
          ws(tapertr+padtraces+1:nx-padtraces-tapertr-1,inode) = source(:,iwa)     ! Source and recv are not padded
          wr(tapertr+padtraces+1:nx-padtraces-tapertr-1,inode) =   recv(:,iwa)
          dws(1,:,inode) = 0.
          dwr(1,:,inode) = 0.
          bws(:,:,inode) = 0.
          bwr(:,:,inode) = 0.
          bws(:,1,inode) = ws(:,inode)
          bwr(:,1,inode) = wr(:,inode)

          ! Background wavefields, bws & bwr
          do iz = 2, img%z%n 
            call etapspi(iwa, inode, -1, ws(:,inode), tflds(:,inode), model(:,iz-1,:), refr(:,iz-1,:),"sou")
            call etapspi(iwa, inode,  1, wr(:,inode), tfldr(:,inode), model(:,iz-1,:), refr(:,iz-1,:),"rec")
            ws(:,inode)=ws(:,inode)*taper
            wr(:,inode)=wr(:,inode)*taper
            bws(:,iz,inode)=ws(:,inode)
            bwr(:,iz,inode)=wr(:,inode)
          end do

          do iz=img%z%n-1, 1, -1
            ! Adjoint imaging condition to get the wavefield perturbation
            call aniimage(.true., iwa, dws(1,:,inode),    bwr(:,iz+1,inode), deltaI(:,:,iz+1),  .true.)
            call aniimage(.true., iwa, bws(:,iz+1,inode), dwr(1,:,inode),    deltaI(:,:,iz+1), .false.)

            ! Upward propagate wavefield perturbations
            call etapspi(iwa, inode, 1, dws(1,:,inode), tflds(:,inode), model(:,iz,:), refr(:,iz,:), "sou")
            dws(1,:,inode)=dws(1,:,inode)*taper
            call etapspi(iwa, inode,-1, dwr(1,:,inode), tfldr(:,inode), model(:,iz,:), refr(:,iz,:), "rec")
            dwr(1,:,inode)=dwr(1,:,inode)*taper

            ! Phase-screen operator & adjoint scattering operator to get the model perturbation
            ddws_s(:,inode)=0; ddws_e(:,inode)=0; 
            call aniphsscr(iwa,inode, 1,ddws_s(:,inode),ddws_e(:,inode),bws(:,iz,inode),tflds(:,inode),model(:,iz,:),refr(:,iz,:),"sou")
            SSS(:,iz,inode)=SSS(:,iz,inode)-conjg(ddws_s(:,inode))*(dws(1,:,inode))/rec%w%n    
            EEE(:,iz,inode)=EEE(:,iz,inode)-conjg(ddws_e(:,inode))*(dws(1,:,inode))/rec%w%n    

            ddwr_s(:,inode)=0; ddwr_e(:,inode)=0;
            call aniphsscr(iwa,inode,-1,ddwr_s(:,inode),ddwr_e(:,inode),bwr(:,iz,inode),tflds(:,inode),model(:,iz,:),refr(:,iz,:),"rec")
            SSS(:,iz,inode)=SSS(:,iz,inode)-conjg(ddwr_s(:,inode))*(dwr(1,:,inode))/rec%w%n    
            EEE(:,iz,inode)=EEE(:,iz,inode)-conjg(ddwr_e(:,inode))*(dwr(1,:,inode))/rec%w%n    
          end do

          !              if (mod(iwa,50)==1) write(3,6) " FREQUENCY # ",iwa," OUT OF ",rec%w%n," / SYNTH.FCT. # ",rec%xd%n, &
          !                   " FROM NODE: ",inode
        end do
        !$OMP END PARALLEL DO

        do inode=1,nodes
          !$OMP parallel do private(ixw,iz)
          do iz=1,nz
            do ixw=1,kxm%n
              deltaM(ixw,iz,1)=deltaM(ixw,iz,1)+SSS(ixw,iz,inode)
              deltaM(ixw,iz,2)=deltaM(ixw,iz,2)+EEE(ixw,iz,inode)
            end do
          end do
          !$OMP END PARALLEL DO
        end do
        deallocate(dwr,dws,EEE,SSS)
        ! ----------------------------------------
        ! Forward tomographic operator
      else 
        allocate(GGG(img%xm%n,nsecdim,img%z%n,nodes))
        GGG=0.
        allocate(dwr(1,kxm%n,nodes));               dwr=0.
        allocate(dws(1,kxm%n,nodes));               dws=0.
        !$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)
        do iwa = 1,rec%w%n
          inode=omp_get_thread_num()+1;
          ws(:,inode)=0.;wr(:,inode)=0.
          ws(tapertr+padtraces+1:nx-padtraces-tapertr-1,inode) = source(:,iwa)
          wr(tapertr+padtraces+1:nx-padtraces-tapertr-1,inode) =   recv(:,iwa)
          dws(1,:,inode)=0.
          dwr(1,:,inode)=0.
          bws(:,:,inode)=0.
          bwr(:,:,inode)=0.
          bws(:,1,inode)=ws(:,inode)
          bwr(:,1,inode)=wr(:,inode)

          do iz = 2, img%z%n 
            call etapspi(iwa, inode, -1, ws(:,inode), tflds(:,inode), model(:,iz-1,:), refr(:,iz-1,:), "sou")
            call etapspi(iwa, inode,  1, wr(:,inode), tfldr(:,inode), model(:,iz-1,:), refr(:,iz-1,:), "rec")
            ws(:,inode)=ws(:,inode)*taper
            wr(:,inode)=wr(:,inode)*taper
            bws(:,iz,inode)=ws(:,inode)
            bwr(:,iz,inode)=wr(:,inode)
          end do

          do iz=1, img%z%n-1
            ! Phase-screen
            ddws_s(:,inode)=0; ddws_e(:,inode)=0; 
            call aniphsscr(iwa,inode,-1,ddws_s(:,inode),ddws_e(:,inode),bws(:,iz,inode),tflds(:,inode),model(:,iz,:),refr(:,iz,:),"sou")
            dws(1,:,inode)=dws(1,:,inode)+(ddws_s(:,inode))*deltaM(:,iz,1)/rec%w%n + (ddws_e(:,inode))*deltaM(:,iz,2)/rec%w%n 
            
            ! Scattering
            call etapspi(iwa, inode, -1, dws(1,:,inode), tflds(:,inode), model(:,iz,:), refr(:,iz,:),"sou")
            dws(1,:,inode)=dws(1,:,inode)*taper

            ! Imaging condition
            call aniimage(.false., iwa, dws(1,:,inode), bwr(:,iz+1,inode), GGG(:,:,iz+1,inode))
!			if(inode==2) write(0,*) "after image", iwa, "inode=", inode, "iz=", iz

            ! Phase-screen
            ddwr_s(:,inode)=0; ddwr_e(:,inode)=0; 
            call aniphsscr(iwa,inode, 1,ddwr_s(:,inode),ddwr_e(:,inode),bwr(:,iz,inode),tflds(:,inode),model(:,iz,:),refr(:,iz,:),"rec")
            dwr(1,:,inode)=dwr(1,:,inode)+(ddwr_s(:,inode))*deltaM(:,iz,1)/rec%w%n + (ddwr_e(:,inode))*deltaM(:,iz,2)/rec%w%n 

            ! Scattering
            call etapspi(iwa, inode,  1, dwr(1,:,inode), tfldr(:,inode), model(:,iz,:), refr(:,iz,:),"rec")
            dwr(1,:,inode)=dwr(1,:,inode)*taper

            ! Imaging condition
            call aniimage(.false., iwa, bws(:,iz+1,inode), dwr(1,:,inode), GGG(:,:,iz+1,inode))
          end do

 !         if (mod(iwa,2)==1) 
!			write(0,6) " FREQUENCY # ",iwa," OUT OF ",rec%w%n," / SYNTH.FCT. # ",rec%xd%n, " FROM NODE: ",inode
        end do
        !$OMP END PARALLEL DO

        do inode=1,nodes
          !$OMP parallel do private(ixw,iz,isecdim)
          do iz=1,img%z%n
            do isecdim=1,nsecdim
              do ixw=1,img%xm%n
                deltaI(ixw,isecdim,iz)=deltaI(ixw,isecdim,iz)+GGG(ixw,isecdim,iz,inode)
              end do
            end do
          end do
          !$OMP END PARALLEL DO
        end do
        deallocate(dwr,dws,GGG)

      end if
    end do
    deallocate(bws,bwr,ws,wr,ddws_s,ddws_e,ddwr_s,ddwr_e,tflds,tfldr)
    deallocate(source, recv, recv_temp, rand_code)
    6   format(A13,i3,A8,i3,A14,i3,A12,i3)

  end subroutine aniwemva_lop_rand
  !======================================================================


  !======================================================================
  subroutine aniwemva_close()

    if (allocated(source1)) deallocate(source1)
    if (allocated(taper))  deallocate(taper)
    if (allocated(model))  deallocate(model)
    if (allocated(refr))   deallocate(refr)

  end subroutine aniwemva_close
  !======================================================================

end module ani_wemvacg2
