module base_trismt_mod
  implicit none
  integer, private			:: n1_s, n2_s, n3_s	

contains

  integer function base_trismt_init(n1_in,n2_in,n3_in) result(stat)
	integer			:: n1_in, n2_in, n3_in

	n1_s = n1_in
	n2_s = n2_in
	n3_s = n3_in
	stat = 0
  end function

  integer function base_tris1d_oper(field,rect1,nsmt,adj) result(stat)
    real         :: field(:)
    integer      :: rect1,nsmt
    integer      :: i1,ismt
    real,pointer :: temp(:)
	logical		 :: adj
    allocate(temp(n1_s))
	if (adj) then
    	do ismt=1, nsmt
       		if (rect1>1 .and. n1_s>1) then
          		call triangle(rect1,1,n1_s,temp,field,adj)
          		field = temp 
       		end if
    	end do
	
	else
    	do ismt=1, nsmt
       		if (rect1>1 .and. n1_s>1) then
          		call triangle(rect1,1,n1_s,field,temp,adj)
          		field = temp 
       		end if
    	end do
	end if
    deallocate(temp)
    stat = 0
  end function

  integer function base_tris2d_oper(field,rect1,rect2,nsmt,adj) result(stat)
    real         :: field(:,:)
    integer      :: rect1,rect2,nsmt
    integer      :: i1,i2,ismt 
    real,pointer :: temp(:)
	logical		 :: adj
    allocate(temp(max(n1_s,n2_s)))  
	if (adj) then
	    do ismt=1, nsmt
       		if (rect1>1 .and. n1_s>1) then
          		do i2=1, n2_s
             		call triangle(rect1,1,n1_s,temp,field(:,i2),adj)
             		field(1:n1_s,i2)=temp(1:n1_s) 
          		end do
       		end if
       		if (rect2>1 .and. n2_s>1) then
          		do i1=1, n1_s
             		call triangle(rect2,1,n2_s,temp,field(i1,:),adj)
             		field(i1,1:n2_s)=temp(1:n2_s)
          		end do
       		end if
    	end do 
	else
	    do ismt=1, nsmt
       		if (rect1>1 .and. n1_s>1) then
          		do i2=1, n2_s
             		call triangle(rect1,1,n1_s,field(:,i2),temp,adj)
             		field(1:n1_s,i2)=temp(1:n1_s) 
          		end do
       		end if
       		if (rect2>1 .and. n2_s>1) then
          		do i1=1, n1_s
             		call triangle(rect2,1,n2_s,field(i1,:),temp,adj)
             		field(i1,1:n2_s)=temp(1:n2_s)
          		end do
       		end if
    	end do 
	end if
    deallocate(temp)
    stat = 0
  end function

  integer function base_tris3d_oper(field,rect1,rect2,rect3,nsmt,adj) result(stat)
    real         :: field(:,:,:)
    integer      :: rect1,rect2,rect3,nsmt
    integer      :: i1,i2,i3,ismt
    real,pointer :: temp(:)
	logical		 :: adj
    allocate(temp(max(n1_s,max(n2_s,n3_s))))
	if (adj) then
   	do ismt=1, nsmt 
       		if (rect1>1 .and. n1_s>1) then
          		do i3=1,n3_s
          		do i2=1,n2_s
             		call triangle(rect1,1,n1_s,temp,field(:,i2,i3),adj)
             		field(1:n1_s,i2,i3)=temp(1:n1_s)
          		end do
         		end do
       		end if
       		if (rect2>1 .and. n2_s>1) then
          		do i3=1,n3_s
          		do i1=1,n1_s
             		call triangle(rect2,1,n2_s,temp,field(i1,:,i3),adj)
             		field(i1,1:n2_s,i3)=temp(1:n2_s)
          		end do
          		end do
       		end if
       		if (rect3>1 .and. n3_s>1) then
          		do i2=1,n2_s
          		do i1=1,n1_s
             		call triangle(rect3,1,n3_s,temp,field(i1,i2,:),adj)
             		field(i1,i2,1:n3_s)=temp(1:n3_s)
          		end do
          		end do
       		end if
    	end do 
	
	else
    	do ismt=1, nsmt 
       		if (rect1>1 .and. n1_s>1) then
          		do i3=1,n3_s
          		do i2=1,n2_s
             		call triangle(rect1,1,n1_s,field(:,i2,i3),temp,adj)
             		field(1:n1_s,i2,i3)=temp(1:n1_s)
          		end do
         		end do
       		end if
       		if (rect2>1 .and. n2_s>1) then
          		do i3=1,n3_s
          		do i1=1,n1_s
             		call triangle(rect2,1,n2_s,field(i1,:,i3),temp,adj)
             		field(i1,1:n2_s,i3)=temp(1:n2_s)
          		end do
          		end do
       		end if
       		if (rect3>1 .and. n3_s>1) then
          		do i2=1,n2_s
          		do i1=1,n1_s
             		call triangle(rect3,1,n3_s,field(i1,i2,:),temp,adj)
             		field(i1,i2,1:n3_s)=temp(1:n3_s)
          		end do
          		end do
       		end if
    	end do 
	end if
    deallocate(temp)
    stat = 0
  end function

  subroutine triangle(nr, m1, n12, uu, vv, adj)
  ! input:        nr      rectangle width (points) (Triangle base twice as wide.)
  ! input:        uu(m1,i2),i2=1,n12      is a vector of data.
  ! output:       vv(m1,i2),i2=1,n12      may be on top of uu
    integer nr,m1,n12, i,np,nq
    real uu( m1, n12),  vv( m1, n12)
    real pp(n12+nr-1), qq(n12+nr+nr-2), tt(n12)
	logical adj

	if (adj) then
	   	do i=1,n12
      		qq(i) = vv(1,i)
    	end do
    	if ( n12 .eq. 1 ) then
    		!  call copy( n12, qq, tt)
       		tt = qq
    	else
       		call boxconv( nr, n12, pp, qq, adj)
       		np = nr+n12-1
       		call boxconv( nr, np , qq, pp, adj)
       		nq = nr+np-1
       		do i= 1, n12
          		tt(i) = qq(i+nr-1)
       		end do
       		do i= 1, nr-1
          		! fold back near end
          		tt(i) = tt(i) + qq(nr-i)
       		end do
       		do i= 1, nr-1
          		! fold back far end
          	tt(n12-i+1) = tt(n12-i+1) + qq(n12+(nr-1)+i)
       		end do
    	end if
    	do i=1,n12
       		uu(1,i) = tt(i)
    	end do
	else
    	do i=1,n12
      		qq(i) = uu(1,i)
    	end do
    	if ( n12 .eq. 1 ) then
    		!  call copy( n12, qq, tt)
       		tt = qq
    	else
       		call boxconv( nr, n12, qq, pp, adj)
       		np = nr+n12-1
       		call boxconv( nr, np , pp, qq, adj)
       		nq = nr+np-1
       		do i= 1, n12
          		tt(i) = qq(i+nr-1)
       		end do
       		do i= 1, nr-1
          		! fold back near end
          		tt(i) = tt(i) + qq(nr-i)
       		end do
       		do i= 1, nr-1
          		! fold back far end
          	tt(n12-i+1) = tt(n12-i+1) + qq(n12+(nr-1)+i)
       		end do
    	end if
    	do i=1,n12
       		vv(1,i) = tt(i)
    	end do
	end if
    return
  end subroutine

  subroutine boxconv(nb, nx, xx, yy, adj)
  ! inputs:       nx,  xx(i), i=1,nx      the data
  !               nb                      the box length
  ! output:       yy(i),i=1,nx+nb-1       smoothed data
    integer nx, ny, nb, i
    real xx(nx), yy(1)
    real bb(nx+nb)
	logical adj
    if ( nb < 1 .or. nb > nx) then
       call erexit('boxconv')  ! "||" means .OR.
    end if
    ny = nx+nb

	if (adj) then   ! adjoint box conv
	    do i= 1, ny
    		bb(i) = 0.
    	end do
    	bb(1) = yy(1)
    	do i= 2, nx
       		bb(i) = bb(i-1) + yy(i)         ! make B(Z) = X(Z)/(1-Z)
    	end do
    	do i= nx+1, ny
       		bb(i) = bb(i-1)
    	end do
    	do i= 1, nb
       		xx(i) = bb(i)
    	end do
    	do i= nb+1, ny
       		xx(i) = bb(i) - bb(i-nb)        ! make Y(Z) = B(Z)*(1-Z**nb)
    	end do
    	do i= 1, ny
       		xx(i) = xx(i) / nb
    	end do
	else           ! forward box conv
	    do i= 1, ny
    		bb(i) = 0.
    	end do
    	bb(1) = xx(1)
    	do i= 2, nx
       		bb(i) = bb(i-1) + xx(i)         ! make B(Z) = X(Z)/(1-Z)
    	end do
    	do i= nx+1, ny
       		bb(i) = bb(i-1)
    	end do
    	do i= 1, nb
       		yy(i) = bb(i)
    	end do
    	do i= nb+1, ny
       		yy(i) = bb(i) - bb(i-nb)        ! make Y(Z) = B(Z)*(1-Z**nb)
    	end do
    	do i= 1, ny
       		yy(i) = yy(i) / nb
    	end do
	end if
    return
  end subroutine


end module

