module eta_phsscr

  use ani_parms_cg

  implicit none

  contains
  !
  ! Anisotropic phase screen
  !
  subroutine aniphsscr_slow(iwa,sign,ddws,wv,slow,eta,dlt)

    complex                 :: ddws(:), wv(:)
    real       	            :: slow(:), eta(:), dlt(:)
    real                    :: dx2, w2, s2, w4, s4, dd, dd2, dd4
    integer                 :: sign,ix,iwa
    complex                 :: iwdz
    complex, allocatable    :: lap2(:), lap4(:)
    real, allocatable       :: a(:), b(:)

    if (.not. allocated(lap2)) allocate(lap2(nx))
    if (.not. allocated(lap4)) allocate(lap4(nx))
    if (.not. allocated(a   )) allocate(   a(nx))
    if (.not. allocated(b   )) allocate(   b(nx))

    a = 0.5*(1.0+2.0*dlt)
    b = 2.0*eta*(1.0+2.0*dlt) + 0.25*(1.0+2.0*dlt)
    lap2=0; lap4=0;

    if (rec%w%all(iwa).ne.0.) then
      iwdz = cmplx(0., 1.)*rec%w%all(iwa)*dz*sign 
      dx2  = dx**2
      w2   = rec%w%all(iwa)**2
      w4   = w2**2
      call laplac2(nx, dx, wv  , lap2)
      call laplac2(nx, dx, lap2, lap4)
      do ix=1, nx 
        dd = sqrt(1+2*dlt(ix)); dd2=dd*dd; dd4=dd2*dd2;
        s2 = slow(ix); s2=s2*s2; s4=s2*s2
        ddws(ix) = dd*iwdz*(wv(ix)*dd+a(ix)/w2/s2/dd2*lap2(ix)+3*a(ix)*b(ix)/w4/s4/dd4*lap4(ix)) 
      end do
    else
      ddws=0.
    end if
    
    if (allocated(lap2)) deallocate(lap2)
    if (allocated(lap4)) deallocate(lap4)
    if (allocated(a)) deallocate(a)
    if (allocated(b)) deallocate(b)

  end subroutine aniphsscr_slow


  subroutine aniphsscr_eta(iwa,sign,ddws,wv,slow,dlt)

    complex                 :: ddws(:), wv(:)
    real	        	:: slow(:),dlt(:)
    real                    :: dx2, w2, s2, w3, s3, dd
    integer                 :: sign,ix,iwa
    complex                 :: iwdz
    complex, allocatable    :: lap2(:), lap4(:)


    if (.not. allocated(lap2)) allocate(lap2(nx))
    if (.not. allocated(lap4)) allocate(lap4(nx))

    lap2=0; lap4=0;

    if (rec%w%all(iwa).ne.0.) then
      iwdz = cmplx(0., 1.)*rec%w%all(iwa)*dz*sign 
      w2   = rec%w%all(iwa)**2
      w3   = w2*rec%w%all(iwa)
      call laplac2(nx, dx, wv  , lap2)
      call laplac2(nx, dx, lap2, lap4)

      do ix=1, kxm%n
        dd = sqrt(1+2*dlt(ix));
        s3 = slow(ix)**3; 
        ddws(ix) = -dd*iwdz/w3/s3*lap4(ix) 
      end do
    else
      ddws=0.
    end if

    if (allocated(lap2)) deallocate(lap2)
    if (allocated(lap4)) deallocate(lap4)
  end subroutine aniphsscr_eta


  !!--------------------------------------------------------------------------
  !! 2nd Order Laplacian

  subroutine laplac2(nnx,ddx, uo,ud)

    integer    :: ix,nnx
    real       :: ddx
    complex, dimension(:)  :: uo,ud

    ud(1)  = (uo(2) - uo(1))/ddx
    ud(nx) = (uo(nx) - uo(nx-1))/ddx
    do ix=1+1,nnx-1
      ud(ix) = &
      (-2.*uo(ix) + uo(ix-1) + uo(ix+1))/dx**2 
    end do
  end subroutine 

end module eta_phsscr
