module killwave_ns

        use ani_parms_cg
        use npolydiv_mod
        use nhelicon_mod
        use wilson_mod
        use nbound_mod
		use nhelix_mod

        implicit none

        real, allocatable, private		:: scale(:)
        real, private					:: ds, os, dang, oang
        integer, private				:: ns, n1_s, n2_s, n3_s
		integer, allocatable,private	:: ptch(:), n_nfilt(:)
		type(nfilter)     ,private 		:: n_filt 
		type( filter)     ,private   	:: hfilt2,lapfilt
		real, private					:: eps

contains 
!######################################
        subroutine  ns_init(n1_in, n2_in, n3_in, dang_in, oang_in)
				integer	:: n1_in, n2_in, n3_in
				real 	:: dang_in, oang_in
                integer	:: i, idip
				real 	:: curdip

				n1_s = n1_in
				n2_s = n2_in
				n3_s = n3_in
				dang = dang_in
				oang = oang_in
				eps  = eps_steer
				write(0,*) "eps", eps, "eps_steer", eps_steer

                allocate (ptch(n1_s*n2_s*n3_s),n_nfilt(nang))
        
                if (lscale)then !scale according to angle
                        call from_param("os",os,1.); call from_param("ds",ds,0.); call from_param("ns",ns,nang)
                        allocate (scale(ns))
                        do i=1,ns
                                scale(i)=os+(i-1)*ds
                        enddo
                endif
        
                ptch=0.
        
                do i=1,n1_s*n2_s*n3_s
                        curdip = atan(dip(i))
                        idip = 1.5+(curdip-oang)/dang
                        ptch(i)=idip
                        !if(ptch(i)>nang.or.(ptch(i)<0.))write(0,*)dip,idip,ptch(i)
                end do
                where (ptch>float(nang))ptch=nang
                where (ptch<1.         )ptch=1
                n_nfilt=nfilt*3
                call nallocate(n_filt,n_nfilt,ptch)
        end subroutine

		subroutine ns_release()
				deallocate(ptch,n_nfilt)
		end subroutine
        !######################################
        subroutine ns_lop(infile, outfile, destory)
				real, intent(in)		:: infile(:)
				real, intent(out)		:: outfile(:)
				logical, intent(in)		:: destory
                real, dimension(2)      :: u,v
				integer					:: iang, ifilt, stat
				real 					:: m, p, lapfilt0, n_filt0
				real				 	:: curdip
                do iang=1,nang

                        do ifilt=1,nfilt
                                n_filt%hlx(iang)%lag(         ifilt) = ifilt			! early lags (first row)
                                n_filt%hlx(iang)%lag( nfilt  +ifilt) = n1_s + ifilt - nfilt 	! late lags  (second row)
                                n_filt%hlx(iang)%lag( nfilt*2+ifilt) = n1_s + ifilt 		! late late lags (second row)
                        end do
                        !call nboundn( iang, (/n1,n2/), (/10, 1/), n_filt)                       !deal with boundary issues here|| Use tentn??
                end do
                call allocatehelix(lapfilt, 4)             ! alloc. filter
                lapfilt0 = 1. + eps                        ! zero lag coeff.
                lapfilt%lag = (/ 1, n1_s-1, n1_s, n1_s+1/)       ! lag(1)= 1; lag(2)=n1  # one side only
        


                do iang=1,nang
                        curdip=oang+(iang-1)*dang
!                        write(0,*)'working on angle',iang,curdip
                        u(1)= cos(curdip); u(2)=-sin(curdip); 
        
!                        write(0,*)'------'
                        m=(u(1)-u(2))/2.; 
                        p=(u(1)+u(2))/2.;
                        lapfilt%flt = 0.;
                        lapfilt%flt = (/2*m*p,-p*p,-2*m*p,-m*m/) 
!                        write(0,*)lapfilt%flt,'|',u,curdip,'m,p',m,p,'nang',nang 
!                        write(0,*)'------'
                        call wilson_init( 10 * n1_s )
!                        write(0,*)'------'
                        call wilson_factor( 100, lapfilt0, lapfilt, n_filt0, n_filt%hlx(iang))
!                        write(0,*)'------'
                        call wilson_close()
                        n_filt%hlx(iang)%flt=n_filt%hlx(iang)%flt!/n_filt0
!                        write(0,*)n_filt%hlx(iang)%flt!/n_filt0
                end do	
        
                if (lscale.and.nang==ns)then    !scale according to the angle!!!!!! useful for zeroing specific angles
                        do iang=1,nang
                                n_filt%hlx(iang)%flt=n_filt%hlx(iang)%flt*scale(iang)
                        end do
                endif
        
                if (destroy)then ! destroy events in n_filt from infile, results in outfile
                        call nhelicon_init(n_filt)
                        stat= filt_nhconest(.false.,.false.,infile,outfile)
                        call nhelicon_close()
                else             ! create events in n_filt from infile, results in outfile
                        call npolydiv_init(n1_s*n2_s*n3_s,n_filt)
                        stat = filt_npolydiv(.false.,.false.,infile,outfile)
                        call npolydiv_close()
                endif
!                write(0,*)'I am writing the output'
!                call srite("outfile", outfile, 4*n1_s*n2_s*n3_s)
        end subroutine
!######################################
        function filt_npolydiv( adj, add, xx, yy) result(stat)
                integer                   :: stat,stat1,stat2 
                logical,intent(in)        :: adj,add 
                real,dimension(:)         :: xx,yy 
                real, dimension(size(xx)) :: tmp
                tmp=0.

                if ( adj) then
                        stat= npolydiv_lop(.false.,.false., yy,tmp)     ! tmp = grad y
                        stat= npolydiv_lop(.true. ,.false., xx,tmp)     ! x = x + grad' tmp
                else
                        stat= npolydiv_lop(.false.,.false., xx,tmp)     ! tmp = grad x
                        stat= npolydiv_lop(.true. ,.false., yy,tmp)     ! y = y + grad' tmp
                endif
                stat=stat*stat2
        end function 
!######################################
        function filt_nhconest( adj, add, xx, yy) result(stat)
                integer                   :: stat,stat1,stat2 
                logical,intent(in)        :: adj,add 
                real,dimension(:)         :: xx,yy 
                real, dimension(size(xx)) :: tmp
                tmp=0.
                if ( adj) then
                        stat= nhelicon_lop(.false.,.false., yy,tmp)     ! tmp = grad y
                        stat= nhelicon_lop(.true. ,.false., xx,tmp)     ! x = x + grad' tmp
                else
                        stat= nhelicon_lop(.false.,.false., xx,tmp)     ! tmp = grad x
                        stat= nhelicon_lop(.true. ,.false., yy,tmp)     ! y = y + grad' tmp
                endif
                stat=stat*stat2
        end function 
end module
